from typing import List, Optional, Union

import numpy as np
import tree  # pip install dm_tree
from ray.rllib.examples.policy.random_policy import RandomPolicy
from ray.rllib.utils.annotations import override
from ray.rllib.utils.typing import TensorStructType, TensorType


class RandomAdmissiblePolicy(RandomPolicy):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @override(RandomPolicy)
    def compute_actions(
        self,
        obs_batch: Union[List[TensorStructType], TensorStructType],
        state_batches: Optional[List[TensorType]] = None,
        prev_action_batch: Union[List[TensorStructType], TensorStructType] = None,
        prev_reward_batch: Union[List[TensorStructType], TensorStructType] = None,
        **kwargs,
    ):
        if "action_mask" in obs_batch.keys():
            action_masks = obs_batch["action_mask"]
            actions = [
                np.random.choice(
                    np.arange(len(action_mask)), p=action_mask / np.sum(action_mask)
                )
                for action_mask in action_masks
            ]
        else:
            obs_batch_size = len(tree.flatten(obs_batch)[0])
            actions = [
                self.action_space_for_sampling.sample() for _ in range(obs_batch_size)
            ]
        return (
            actions,
            [],
            {},
        )



from gobblet_rl.game.gobblet import env, parallel_env, raw_env  # noqa: 401
from gobblet_rl.game.greedy_policy import GreedyGobbletPolicy  # noqa 401
from gobblet_rl.game.manual_policy import ManualGobbletPolicy  # noqa 401



import argparse
import time

import numpy as np

from gobblet_rl import gobblet_v1


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--render_mode",
        type=str,
        default="human",
        choices=["human", "rgb_array", "text", "text_full"],
        help="Choose the rendering mode for the game.",
    )

    parser.add_argument(
        "--seed", type=int, default=None, help="random seed for board and policy"
    )
    parser.add_argument(
        "--debug", action="store_true", help="display extra debugging information"
    )
    parser.add_argument(
        "--screen-width", type=int, default=640, help="Width of pygame screen in pixels"
    )

    return parser


def get_args() -> argparse.Namespace:
    parser = get_parser()
    return parser.parse_known_args()[0]


if __name__ == "__main__":
    # train the agent and watch its performance in a match!
    args = get_args()

    if args.seed is not None:
        np.random.seed(args.seed)

    env = gobblet_v1.env(render_mode=args.render_mode, args=args)

    env.reset()

    env.render()  # need to render the environment before pygame can take user input

    for agent in env.agent_iter():
        observation, reward, termination, truncation, info = env.last()

        if termination or truncation:
            print(f"Agent: ({agent}), Reward: {reward}, info: {info}")
            env.step(None)

        else:
            action_mask = observation["action_mask"]
            action = np.random.choice(
                np.arange(len(action_mask)), p=action_mask / np.sum(action_mask)
            )

            if args.render_mode == "human":
                # Wait .5 seconds between moves so the user can follow the sequence of moves
                time.sleep(0.5)

            env.step(action)



# adapted from pettingzoo.classic.connect_four
import glob
import os
import re
import subprocess
import time
from pathlib import Path
from typing import Union

import pygame


def get_image(path):
    """Get image.

    Load an image from file into a pygame object

    Returns:
        sfc: Pygame surface object with the image
    """
    cwd = os.path.dirname(__file__)
    image = pygame.image.load(os.path.join(cwd, path))
    sfc = pygame.Surface(image.get_size(), flags=pygame.SRCALPHA)
    sfc.blit(image, (0, 0))
    return sfc


def load_chip(tile_size, filename, scale):
    """Load chip.

    Load the image of a single chip

    Returns:
        Chip: Pygame surface object containing the chip
    """
    chip = get_image(os.path.join("img", filename))
    chip = pygame.transform.scale(
        chip, (int(tile_size * (scale)), int(tile_size * (scale)))
    )
    return chip


def load_chip_preview(tile_size, filename, scale):
    """Load chip preview.

    Load the preview image of a single chip

    Returns:
        Chip: Pygame surface object containing the chip preview
    """
    chip = get_image(os.path.join(os.path.join("img", "preview"), filename))
    chip = pygame.transform.scale(
        chip, (int(tile_size * (scale)), int(tile_size * (scale)))
    )
    return chip


# from https://github.com/michaelfeil/skyjo_rl/blob/dev/rlskyjo/utils.py


def get_project_root() -> Path:
    """Get project root.

    return Path to the project directory, top folder of gobblet-rl.

    Returns:
        Path: Path to the project directory
    """
    return Path(__file__).parent.parent.parent.resolve()


def find_file_in_subdir(
    parent_dir: Union[Path, str], file_str: Union[Path, str], regex_match: str = None
) -> Union[str, None]:
    """Find file in subdirectory.

    Finds the path of a file in a subdirectory

    Args:
        parent_dir: directory to look for the file in

    Returns:
        Path: Path to the project directory
    """
    files = glob.glob(os.path.join(parent_dir, "**", file_str), recursive=True)
    if regex_match is not None:
        p = re.compile(regex_match)
        files = [s for s in files if p.match(s)]
    return sorted(files)[-1] if len(files) else None


# adapted from https://commons.wikimedia.org/wiki/File:SquareWaveFourierArrows.gif#Source_code
class GIFRecorder:
    """GIF Recorder.

    This class is used to record a PyGame surface and save it to a .gif file.
    """

    def __init__(self, out_file="game.gif"):
        """Initialize the recorder.

        Args:
            out_file: Output file to save the recording
        """
        print("Initializing GIF Recorder...")
        print(f"Output of the recording will be saved to {out_file}.")
        self.filename_list = []
        self.frame_num = 0
        self.start_time = time.time()
        self.path = get_project_root()
        print(self.path)
        self.out_file = out_file
        self.ended = False

    def capture_frame(self, surf):
        """Capture frame.

         Call this method every frame, pass in the pygame surface to capture.
         Note: surface must have the dimensions specified in the constructor.

        Args:
            surf: pygame surface to capture

        Returns:
             None
        """
        if not self.ended:  # Stop saving frames after we have exported the recording
            # transform the pixels to the format used by open-cv
            self.filename_list.append(
                os.path.join(
                    self.path, f"temp_{time.time()}_" + str(self.frame_num) + ".png"
                )
            )
            pygame.image.save(surf, self.filename_list[-1])
            self.frame_num += 1

    # Convert individual image files to GIF
    def end_recording(self, surf):
        """End recording.

        Call this method to stop recording.

        :return: None
        """
        if not self.ended:
            # Add 10 frames to the end of the video so people have a chance to see the move
            for i in range(10):
                self.capture_frame(surf)

            # stop recording
            duration = time.time() - self.start_time
            seconds_per_frame = duration / self.frame_num
            frame_delay = str(int(seconds_per_frame * 100))
            command_list = (
                ["convert", "-delay", frame_delay, "-loop", "0"]
                + self.filename_list
                + [self.out_file]
            )
            # Use the "convert" command (part of ImageMagick) to build the animation
            subprocess.call(command_list, cwd=self.path)
            # Earlier, we saved an image file for each frame of the animation. Now
            # that the animation is assembled, we can (optionally) delete those files
            for filename in self.filename_list:
                os.remove(filename)
            print(f"Saved recording to {self.out_file}")
            self.ended = True



from typing import Any, Optional

import numpy as np

from gobblet_rl.game.board import Board


class GreedyGobbletPolicy:
    def __init__(
        self,
        depth: Optional[int] = 2,
        seed: Optional[int] = 0,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.board = None
        self.depth = depth
        self.rng = np.random.default_rng()
        self.prev_actions = {i: [] for i in range(2)}

    def compute_actions_rllib(self, obs_batch):
        observations = obs_batch["observation"]
        observations = observations.reshape(
            observations.shape[0], 3, 3, -1
        )  # Infer observation dimension and batch size
        masks = obs_batch["action_mask"]
        actions = []
        for i in range(len(observations)):
            act = self.compute_action(observations[i], masks[i])
            actions.append(act)
        return actions

    def compute_action_tianshou(self, obs):
        mask = obs.mask
        obs = obs.obs if hasattr(obs, "obs") else obs  #
        return self.compute_action(obs, mask)

    def compute_action(
        self,
        obs,
        mask,
    ) -> np.ndarray:
        obs_player = obs[..., :6]  # Index the last dimension
        board_player = np.array(
            [
                (i + 1) * obs_player[..., i] + (i + 2) * obs_player[..., i + 1]
                for i in range(0, 6, 2)
            ]
        )  # Reshapes [3,3,6] to [3,3,3]
        obs_opponent = obs[..., 6:12]
        board_opponent = np.array(
            [
                (i + 1) * obs_opponent[..., i] + (i + 2) * obs_opponent[..., i + 1]
                for i in range(0, 6, 2)
            ]
        )
        board = np.where(board_player > board_opponent, board_player, -board_opponent)

        agent_index = obs[
            ..., 12
        ].max()  # Thirteenth layer of obs encodes agent_index (all zeros or all ones)
        opponent_index = (
            1 - agent_index
        )  # If agent index is 1, we want 0; if agent index is 0, we want 1

        # If we are playing as the second agent, we need to adjust the representation of the board to reflect that
        if agent_index == 1:
            board = -board

        self.board = Board()
        self.board.squares = board.flatten().copy()

        # Map agent index to winner value that the board will return (agent idx: [0,1], winner_vals: [1,-1])
        winner_values = [1, -1]

        legal_actions = mask.flatten().nonzero()[0]
        actions_depth1 = list(
            legal_actions
        )  # Initialize the same as legal actions, then remove ones that cause a loss
        chosen_action = None

        results = {}
        # Depth 1
        for action in legal_actions:
            if self.board.is_legal(agent_index=agent_index, action=action):
                depth1_board = Board()
                depth1_board.squares = self.board.squares.copy()
                depth1_board.play_turn(agent_index=agent_index, action=action)

                # If we can win next turn, do it
                results[action] = depth1_board.check_for_winner()
                if results[action] == winner_values[agent_index]:  # Win for our agent
                    chosen_action = action
                    break
                elif (
                    results[action] == winner_values[opponent_index]
                ):  # Loss for our agent
                    if len(actions_depth1) > 1:
                        actions_depth1.remove(action)
                    else:
                        break  # If there is nothing we can do to prevent them from winning, we have to do one of the moves

        if self.depth > 1:
            # Depth 2 (only explore cases where the immediate action doesn't result in a win or loss)
            for action in [key for key, value in results.items() if value == 0]:
                # if self.board.is_legal(agent_index=agent_index, action=action):
                depth1_board = Board()
                depth1_board.squares = self.board.squares.copy()
                depth1_board.play_turn(agent_index=agent_index, action=action)

                # Check what the opponent can do after this potential move
                legal_actions_depth2 = [
                    act
                    for act in range(len(mask.flatten()))
                    if depth1_board.is_legal(agent_index=opponent_index, action=act)
                ]

                results_depth2 = {}
                for action_depth2 in legal_actions_depth2:
                    depth2_board = Board()
                    depth2_board.squares = depth1_board.squares.copy()
                    depth2_board.play_turn(
                        agent_index=opponent_index, action=action_depth2
                    )

                    results_depth2[action_depth2] = depth2_board.check_for_winner()

                    # If the opponent can win in the next move, we don't want to do this action
                    if (
                        results_depth2[action_depth2] == winner_values[opponent_index]
                    ):  # Check if it's possible for the opponent to win next turn after
                        if len(actions_depth1) > 1:
                            if action in actions_depth1:
                                actions_depth1.remove(action)
                        else:
                            break  # If there is nothing we can do to prevent them from winning, we have to pick one

                        # If we can put our piece in the place that the opponent would have gone to win the game, do that
                        # BUT this might miss us from winning the game ourselves in blcoking them, so don't exit the loop
                        # So we only do it if there aren't deterministic wins already chosen (and we don't break, keep looking after)
                        if self.board.is_legal(action_depth2, agent_index=agent_index):
                            if chosen_action is None:
                                chosen_action = action_depth2

                # Depth 2: if this move sets the opponent up with only moves that win the game for us, pick this move
                if all(
                    winner == winner_values[agent_index]
                    for winner in results_depth2.values()
                ):
                    chosen_action = action
                    break
                # Depth 2: if this move blocks the opponent from winning, pick this if we cannot find any guaranteed wins
                if all(
                    winner != winner_values[opponent_index]
                    for winner in results_depth2.values()
                ):
                    chosen_action = action  # BLOCKING ACTION

                    # Depth 3: Given that we block the opponent this way, can we win the next turn?
                    if self.depth == 3:
                        depth1_board = Board()
                        depth1_board.squares = self.board.squares.copy()
                        depth1_board.play_turn(agent_index=agent_index, action=action)

                        # Search over depth 2 actions where the opponent doesn't win
                        for action_depth2 in [
                            key for key, value in results_depth2.items() if value == 0
                        ]:
                            depth2_board = Board()
                            depth2_board.squares = depth1_board.squares.copy()
                            depth2_board.play_turn(
                                agent_index=agent_index, action=action_depth2
                            )

                            legal_actions_depth3 = [
                                act
                                for act in range(len(mask.flatten()))
                                if depth2_board.is_legal(
                                    agent_index=agent_index, action=act
                                )
                            ]
                            actions_depth3 = list(legal_actions_depth3)
                            res_depth3 = {}

                            # Search over all depth 3 actions which we can do
                            for act_depth3 in legal_actions_depth3:
                                depth3_board = Board()
                                depth3_board.squares = depth2_board.squares.copy()
                                depth3_board.play_turn(
                                    agent_index=agent_index, action=action
                                )

                                # If we can win next turn, do it
                                res_depth3[act_depth3] = depth3_board.check_for_winner()
                                if (
                                    res_depth3[act_depth3] == winner_values[agent_index]
                                ):  # Win for our agent
                                    chosen_action = action  # If we can win in depth 3, then we know this blocking action is good
                                    break
                                elif (
                                    res_depth3[act_depth3]
                                    == winner_values[opponent_index]
                                ):  # Loss for our agent
                                    if len(actions_depth3) > 1:
                                        if action in actions_depth3:
                                            actions_depth3.remove(action)
                                    else:
                                        break  # If there is nothing we can do to prevent them from winning in depth3, we have to do one of the moves

        # If we have not selected an action, or have already done it in the last 3 turns, choose randomly
        if (
            chosen_action is None
            or chosen_action in self.prev_actions[agent_index][-3:]
        ):
            # Choose randomly between possible actions:
            # chosen_action = self.rng.choice(actions_depth1)
            chosen_action = np.random.choice(actions_depth1)
            # print(f"Choosing randomly between possible actions: {actions_depth1} --> {chosen_action}")
        self.prev_actions[agent_index].append(chosen_action)
        act = np.array(chosen_action)
        return act



import gymnasium
import gymnasium.core
import numpy as np
from pettingzoo.utils import wrappers
from ray.rllib.agents import ppo
from ray.rllib.env import ParallelPettingZooEnv
from ray.rllib.policy.policy import PolicySpec
from ray.tune import register_env

from gobblet_rl import gobblet_v1

# # Note: gymnasium.core.Wrapper will cause the observation to be a string 'player_1', rather than the dict
# class ReshapedEnv(gym.core.Wrapper):
#     def __init__(self, env):
#         super().__init__(env)
#
#         # Number of actions for each agent (for redefining action space method below)
#         self.n = {}
#
#         # Flatten observation (RLlib will change model type from fully connected to CNN for obs space with 3+ dims)
#         for p in env.possible_agents:
#             old = self.observation_space(p)['observation']
#             new_shape = (old.shape[0] * old.shape[1], old.shape[2])
#             self.observation_space(p)['observation'] = gymnasium.spaces.Box(low=old.low.reshape(new_shape), high=old.high.reshape(new_shape), shape=new_shape, dtype=old.dtype)
#
#             # Number of actions for this agent
#             self.n[p] = env.action_space(env.possible_agents[0]).n
#
#     # Convert to gym.spaces.Discrete (RLlib model catalog expects action space to be gym.spaces.Discrete. rather than gymnasium)
#     def action_space(self, agent):
#         return gym.spaces.Discrete(self.n[agent])
#
#     # Override `observation` to custom process the original observation
#     # coming from the env.
#     def observation(self, observation):
#         # If the observation consists of a dict with keys ['player_1', 'player_2'] (expected for parallel env)
#         if isinstance(observation, dict):
#             if self.env.agents ==  list(observation.keys()):
#                 for key in observation.keys():
#                     observation[key]["observation"] = observation[key]["observation"].reshape(9, 12)
#             elif "observation" in observation.keys():
#                 observation["observation"] = observation["observation"].reshape(9, 12)
#         else:
#             observation = observation.reshape(9, 12)
#         return observation


class FlattenedEnv(wrappers.base.BaseWrapper):
    def __init__(self, env):
        # # Number of actions for each agent (for redefining action space method below)
        # self.n = {}
        super().__init__(env)

        # Flatten observation (RLlib will change model type from fully connected to CNN for obs space with 3+ dims)
        for p in env.possible_agents:
            old = self.observation_space(p)["observation"]
            new_shape = np.prod(old.shape).reshape(1)
            self.observation_space(p)["observation"] = gymnasium.spaces.Box(
                low=old.low.reshape(new_shape),
                high=old.high.reshape(new_shape),
                shape=new_shape,
                dtype=old.dtype,
            )

    def observe(self, agent):
        observation = self.env.observe(agent)
        print("observation: ", observation)

        new_shape = self.observation_space(self.env.possible_agents[0])[
            "observation"
        ].shape
        # print(new_shape)
        # If the observation consists of a dict with keys ['player_1', 'player_2'] (expected for parallel env)

        if isinstance(observation, dict):
            if self.env.agents == list(observation.keys()):
                for key in observation.keys():
                    observation[key]["observation"] = observation[key][
                        "observation"
                    ].reshape(new_shape)
            elif "observation" in observation.keys():
                observation["observation"] = observation["observation"].reshape(
                    new_shape
                )
        else:
            observation = observation.reshape(new_shape)
        print("flattened observation: ", observation)
        return observation

    def __str__(self):
        return str(self.env)


def env_creator(args):
    env = gobblet_v1.parallel_env()
    env = FlattenedEnv(env)
    # env = ReshapedEnv(env)
    # env = ss.reshape_v0(env, (9, 12))
    return env


if __name__ == "__main__":
    env_name = "gobblet_pettingzoo"

    register_env(env_name, lambda config: ParallelPettingZooEnv(env_creator(config)))

    test_env = ParallelPettingZooEnv(env_creator({}))

    # # Convert obs space and act space to gym (what RLlib expects)
    # obs_space = test_env.observation_space
    # obs_space = obs_space["observation"]
    # obs_space = gym.spaces.Box(obs_space.low, obs_space.high, obs_space.shape, obs_space.dtype)
    # act_space = test_env.action_space
    # act_space = gym.spaces.Discrete(act_space.n)

    agents = ["player_1", "player_2"]
    custom_config = {
        "env": env_name,
        "framework": "torch",
        # Use GPUs iff `RLLIB_NUM_GPUS` env var set to > 0.
        # "num_gpus": int(torch.cuda.device_count()),
        "num_gpus": 0,
        "num_workers": 1,  # os.cpu_count() - 1,
        "multiagent": {
            "policies": {
                name: (
                    PolicySpec(
                        policy_class=None,  # infer automatically from Algorithm
                        observation_space=test_env.observation_space,  # infer automatically from env
                        action_space=test_env.action_space,  # infer automatically from env
                        config={
                            "gamma": 0.85
                        },  # use main config plus <- this override here
                    )
                )
                for name in agents
            },
            "policy_mapping_fn": lambda agent_id: agent_id,
        },
    }

    ppo_config = ppo.DEFAULT_CONFIG.copy()
    ppo_config.update(custom_config)

    trainer = ppo.PPOTrainer(config=ppo_config)

    result = trainer.train()

    #  # Below code might define policy wrong, was copied from somewhere else
    #
    # def gen_policy(i):
    #     config = {
    #         "gamma": 0.99,
    #     }
    #     return (None, test_env.observation_space, test_env.action_space, config)
    #
    # policies = {"policy_0": gen_policy(0)}
    #
    # policy_ids = list(policies.keys())
    #
    # tune.run(
    #     "PPO",
    #     name="PPO",
    #     stop={"timesteps_total": 5000000},
    #     checkpoint_freq=10,
    #     local_dir="~/ray_results/"+env_name,
    #     config={
    #         # Environment specific
    #         "env": env_name,
    #         "no_done_at_end": True,
    #         "num_gpus": 0,
    #         "num_workers": 2,
    #         "num_envs_per_worker": 1,
    #         "compress_observations": False,
    #         "batch_mode": 'truncate_episodes',
    #         "clip_rewards": False,
    #         "vf_clip_param": 500.0,
    #         "entropy_coeff": 0.01,
    #         # effective batch_size: train_batch_size * num_agents_in_each_environment [5, 10]
    #         # see https://github.com/ray-project/ray/issues/4628
    #         "train_batch_size": 1000,  # 5000
    #         "rollout_fragment_length": 50,  # 100
    #         "sgd_minibatch_size": 100,  # 500
    #         "vf_share_layers": False
    #         },
    # )



from typing import List, Optional, Union

import numpy as np
from ray.rllib.examples.policy.random_policy import RandomPolicy
from ray.rllib.utils.annotations import override
from ray.rllib.utils.typing import TensorStructType, TensorType

from gobblet_rl.game.greedy_policy import GreedyGobbletPolicy


class GreedyPolicy(RandomPolicy):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.policy = GreedyGobbletPolicy(seed=np.random.randint(1000), depth=1)

    @override(RandomPolicy)
    def compute_actions(
        self,
        obs_batch: Union[List[TensorStructType], TensorStructType],
        state_batches: Optional[List[TensorType]] = None,
        prev_action_batch: Union[List[TensorStructType], TensorStructType] = None,
        prev_reward_batch: Union[List[TensorStructType], TensorStructType] = None,
        **kwargs,
    ):
        actions = self.policy.compute_actions_rllib(obs_batch)
        return (
            actions,
            [],
            {},
        )



# Adapted from https://github.com/Farama-Foundation/PettingZoo/blob/master/pettingzoo/butterfly/knights_archers_zombies/manual_policy.py
import sys

import numpy as np
import pygame

from .utils import GIFRecorder


class ManualGobbletPolicy:
    def __init__(self, env, agent_id: int = 0, recorder: GIFRecorder = None):
        self.env = env
        self.agent_id = agent_id
        self.agent = self.env.agents[self.agent_id]
        self.recorder = recorder

        env.render()  # need to render the environment before pygame can take user input

    def __call__(self, observation, agent):
        recorder = self.recorder
        env = self.env

        picked_up = False
        picked_up_pos = -1
        piece_cycle = 0
        piece_size_selected = 0

        while True:
            event = pygame.event.wait()

            if event.type == pygame.QUIT:
                if recorder is not None:
                    recorder.end_recording(env.unwrapped.screen)
                pygame.quit()
                pygame.display.quit()
                sys.exit()

            """ GET MOUSE INPUT"""
            mousex, mousey = pygame.mouse.get_pos()
            width, height = pygame.display.get_surface().get_size()
            pos_x = 0
            if 0 * width / 1000 <= mousex < 360 * width / 1000:
                pos_x = 0
            elif 360 * width / 1000 <= mousex < 640 * width / 1000:
                pos_x = 1
            elif 640 * width / 1000 <= mousex < 1000 * width / 1000:
                pos_x = 2
            pos_y = 0
            if 0 * height / 1000 <= mousey < 360 * height / 1000:
                pos_y = 0
            elif 360 * height / 1000 <= mousey < 640 * height / 1000:
                pos_y = 1
            elif 640 * height / 1000 <= mousey < 1000 * height / 1000:
                pos_y = 2
            pos = pos_y + 3 * (pos_x)

            agent_multiplier = 1 if agent == env.agents[0] else -1

            """ FIND PLACED PIECES """
            placed_pieces = env.unwrapped.board.squares[
                env.unwrapped.board.squares.nonzero()
            ]
            placed_pieces_agent = [
                a for a in placed_pieces if np.sign(a) == agent_multiplier
            ]
            placed_pieces_agent_abs = [abs(p) for p in placed_pieces_agent]
            pieces = np.arange(1, 7)
            unplaced = [p for p in pieces if p not in placed_pieces_agent_abs]
            flatboard = env.unwrapped.board.get_flatboard()

            # Selected piece size persists as we loop through events, allowing the user to cycle through sizes
            if piece_size_selected == 0:
                if len(unplaced) > 0:
                    piece = unplaced[-1]
                    piece_size_selected = (piece + 1) // 2
                else:
                    piece = -1

            """ READ KEYBOARD INPUT"""
            if event.type == pygame.KEYDOWN:
                if not picked_up:
                    if (
                        event.key == pygame.K_SPACE
                    ):  # Cycle through pieces (from largest to smallest)
                        piece_cycle += 1

                        cycle_choices = np.unique(
                            [(p + 1) // 2 for p in unplaced]
                        )  # Transform [1,2,3,4,5,6] to [1,2,3)

                        if len(cycle_choices) > 0:
                            piece_size = cycle_choices[
                                (np.amax(cycle_choices) - (piece_cycle + 1))
                                % len(cycle_choices)
                            ]
                        # else:
                        #   piece_size = -1
                        piece_size_selected = piece_size

                        if (
                            piece_size * 2
                        ) - 1 in unplaced:  # Check if the first of this piece size is available
                            piece = piece_size * 2 - 1
                        else:
                            piece = (
                                piece_size * 2
                            )  # Otherwise choose the second of this piece size
                    else:
                        if event.key == pygame.K_1:  # Select piece size 1
                            piece_size_selected = 1
                            if 1 in unplaced:
                                piece = 1
                                piece_cycle = 2
                            elif 2 in unplaced:
                                piece = 2
                                piece_cycle = 2
                            else:
                                piece = -1
                        elif event.key == pygame.K_2:  # Select piece size 2
                            piece_size_selected = 2
                            if 3 in unplaced:
                                piece = 3
                                piece_cycle = 1
                            elif 4 in unplaced:
                                piece = 4
                                piece_cycle = 1
                            else:
                                piece = -1
                        elif event.key == pygame.K_3:  # Select piece size 3
                            piece_size_selected = 3
                            if 5 in unplaced:
                                piece = 5
                                piece_cycle = 0
                            elif 6 in unplaced:
                                piece = 6
                                piece_cycle = 0
                            else:
                                piece = -1

            # Don't render a preview if both pieces of a given size have been placed
            if piece != -1:
                piece_size = (piece + 1) // 2
                # Don't render a preview if both pieces of a given size have been placed

                """ GET PREVIEW ACTION """
                # Get the action from the preview (in position the mouse cursor is currently hovering over)
                action_prev = env.unwrapped.board.get_action(
                    pos, piece_size, env.agents.index(agent)
                )

            """ CLEAR ACTION PREVIEW FOR ILLEGAL MOVES"""
            # If the hovered over position means placing a picked up piece in the same spot, mark it as illegal
            if pos == picked_up_pos or piece == -1:
                action_prev = -1

            """ CLEAR PREVIOUSLY PREVIEWED MOVES """
            env.unwrapped.board.squares_preview[:] = 0
            if action_prev != -1:
                if not env.unwrapped.board.is_legal(
                    action_prev, env.agents.index(agent)
                ):  # If this action is illegal
                    action_prev = -1
                else:
                    env.unwrapped.board.squares_preview[
                        pos + 9 * (piece_size - 1)
                    ] = agent_multiplier  # Preview this position

            """ UPDATE DISPLAY with previewed move"""
            env.render()
            pygame.display.update()
            if recorder is not None:
                recorder.capture_frame(env.unwrapped.screen)

            """ PICK UP / PLACE A PIECE """
            if event.type == pygame.MOUSEBUTTONDOWN:
                # Pick up a piece (only able to if it has already been placed, and is not currently picked up)
                if flatboard[pos] in placed_pieces_agent and not picked_up:
                    piece_size_on_board = (abs(flatboard[pos]) + 1) // 2
                    piece_to_pick_up = int(flatboard[pos])
                    piece = abs(piece_to_pick_up)

                    # If the piece size selected is larger, clicking here should self-gobble the smaller piece
                    if piece_size_on_board >= piece_size_selected:
                        # Can only pick up a piece if there is a legal move to place it, other than where it was before
                        if not all(
                            observation["action_mask"][9 * (piece - 1) : 9 * piece] == 0
                        ):
                            picked_up = True
                            picked_up_pos = pos
                            # Remove a placed piece
                            piece_size_selected = (piece + 1) // 2

                            index = np.where(
                                env.unwrapped.board.squares == piece_to_pick_up
                            )[0][0]
                            env.unwrapped.board.squares[index] = 0

                            # Set the only possible actions to be moving this piece to a new square
                            observation["action_mask"][
                                pos + 9 * (piece - 1)
                            ] = 0  # TODO: check if this is already zero
                            observation["action_mask"][
                                : 9 * (piece - 1)
                            ] = 0  # Zero out all the possible actions
                            observation["action_mask"][9 * (piece) :] = 0

                # Place a piece (if it is legal to do so)
                else:
                    if action_prev != -1:
                        env.unwrapped.board.squares_preview[
                            pos + 9 * (piece_size - 1)
                        ] = 0
                        action = pos + 9 * (piece - 1)
                        break
        return np.int32(action)

    @property
    def available_agents(self):
        return self.env.agent_name_mapping



from copy import deepcopy
from typing import Any, Dict, Optional, Union

import numpy as np
from numpy import ndarray
from tianshou.data import Batch
from tianshou.policy import BasePolicy

from gobblet_rl.game.greedy_policy import GreedyGobbletPolicy


class GreedyPolicy(BasePolicy):
    """Greedy Policy.

    Basic greedy policy which checks if a move results in a victory, and if it sets the opponent up to win (or lose) in the next turn.
    The depth argument controls the agent's search depth (default of 2 is a balance between computational efficiency and optimal play)
     * depth = 1: Agent considers moves which it can use to directly win
     * depth = 2: Agent also considers moves it can take to block the opponent from winning next turn
     * depth = 3: Agent also considers moves which set it up to win in two moves: no matter what opponents does in retaliation (unblockable wins)
    """

    def __init__(
        self,
        depth: Optional[int] = 2,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.board = None
        self.depth = depth
        self.policy = GreedyGobbletPolicy()

    def forward(
        self,
        batch: Batch,
        state: Optional[Union[dict, Batch, np.ndarray]] = None,
        input: str = "obs",
        **kwargs: Any,
    ) -> Batch:
        """Compute action over the given batch data.

        If you need to mask the action, please add a "mask" into batch.obs, for
        example, if we have an environment that has "0/1/2" three actions:
        ::

            batch == Batch(
                obs=Batch(
                    obs="original obs, with batch_size=1 for demonstration",
                    mask=np.array([[False, True, False]]),
                    # action 1 is available
                    # action 0 and 2 are unavailable
                ),
                ...
            )

        :return: A :class:`~tianshou.data.Batch` with "act" key, containing
            the greedy action.

        .. seealso::

            Please refer to :meth:`~tianshou.policy.BasePolicy.forward` for
            more detailed explanation.
        """
        full_batch = deepcopy(batch)
        if len(batch) > 1:
            acts = []
            for b in range(len(batch)):
                batch_single = deepcopy(full_batch)
                if len(batch_single[input].agent_id) > 1:
                    batch_single[input].obs = batch_single[input].obs[b, ...]
                    batch_single[input].mask = batch_single[input].mask[b, ...]
                    batch_single[input].agent_id = (
                        batch_single[input].agent_id[b, ...].reshape(1)
                    )
                act = self.forward_unbatched(batch=batch_single)  # array(1)
                acts.append(act)  # [ array(1), array(2), ... array(10) ]
            acts = np.array(acts)
            return Batch(
                act=acts
            )  # Batch( act: [ array(1), array(2), ... array(10) ] )
        else:
            batch_single = deepcopy(full_batch)
            act = self.forward_unbatched(batch=batch_single)
            acts = np.array(act).reshape(1)
            return Batch(act=acts)

    def forward_unbatched(
        self,
        batch: Batch,
        input: str = "obs",
        **kwargs: Any,
    ) -> ndarray:
        obs = batch[input]
        act = self.policy.compute_action_tianshou(obs)
        return act

    def learn(self, batch: Batch, **kwargs: Any) -> Dict[str, float]:
        """Since a random agent learns nothing, it returns an empty dict."""
        return {}



import argparse

import numpy as np
import pygame


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--seed",
        type=int,
        default=None,
        help="Set random seed manually (will only affect CPU agents)",
    )
    parser.add_argument(
        "--cpu-players",
        type=int,
        default=1,
        choices=[0, 1, 2],
        help="Number of CPU players (options: 0, 1, 2)",
    )
    parser.add_argument(
        "--player",
        type=int,
        default=0,
        choices=[0, 1],
        help="Choose which player to play as: red = 0, yellow = 1",
    )
    parser.add_argument(
        "--screen-width", type=int, default=640, help="Width of pygame screen in pixels"
    )
    return parser


def get_args() -> argparse.Namespace:
    parser = get_parser()
    return parser.parse_known_args()[0]


if __name__ == "__main__":
    from gobblet_rl import gobblet_v1

    args = get_args()

    clock = pygame.time.Clock()

    if args.seed is not None:
        np.random.seed(args.seed)

    env = gobblet_v1.env(render_mode="human", args=args)
    env.reset()

    manual_policy = gobblet_v1.ManualGobbletPolicy(env)

    for agent in env.agent_iter():
        clock.tick(env.metadata["render_fps"])

        observation, reward, termination, truncation, info = env.last()

        if termination or truncation:
            print(f"Agent: ({agent}), Reward: {reward}, info: {info}")
            env.step(None)
            continue

        if agent == manual_policy.agent and args.cpu_players < 2:
            action = manual_policy(observation, agent)
        else:
            action_mask = observation["action_mask"]
            action = np.random.choice(
                np.arange(len(action_mask)), p=action_mask / np.sum(action_mask)
            )

        env.step(action)

        env.render()



# adapted from https://github.com/Farama-Foundation/PettingZoo/blob/master/tutorials/Tianshou/3_cli_and_logging.py

import argparse
import os
import time
from copy import deepcopy
from typing import Optional, Tuple

import gym
import gymnasium
import numpy as np
import torch
from tianshou.data import Collector, VectorReplayBuffer
from tianshou.env import DummyVectorEnv
from tianshou.env.pettingzoo_env import PettingZooEnv
from tianshou.policy import BasePolicy, DQNPolicy, MultiAgentPolicyManager
from tianshou.trainer import offpolicy_trainer
from tianshou.utils import TensorboardLogger
from tianshou.utils.net.common import Net
from torch.utils.tensorboard import SummaryWriter

from gobblet_rl import gobblet_v1
from gobblet_rl.game.collector_manual_policy import ManualPolicyCollector
from gobblet_rl.game.greedy_policy_tianshou import GreedyPolicy
from gobblet_rl.game.utils import GIFRecorder


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument("--seed", type=int, default=1626)
    parser.add_argument("--eps-test", type=float, default=0.05)
    parser.add_argument("--eps-train", type=float, default=0.1)
    parser.add_argument("--buffer-size", type=int, default=20000)
    parser.add_argument(
        "--lr", type=float, default=1e-4
    )  # TODO: Changing this to 1e-5 for some reason makes it pause after 3 or 4 epochs
    parser.add_argument(
        "--gamma", type=float, default=0.9, help="a smaller gamma favors earlier win"
    )
    parser.add_argument("--n-step", type=int, default=3)
    parser.add_argument("--target-update-freq", type=int, default=320)
    parser.add_argument("--epoch", type=int, default=50)
    parser.add_argument("--step-per-epoch", type=int, default=1000)
    parser.add_argument("--step-per-collect", type=int, default=10)
    parser.add_argument("--update-per-step", type=float, default=0.1)
    parser.add_argument("--batch-size", type=int, default=64)
    parser.add_argument(
        "--hidden-sizes", type=int, nargs="*", default=[128, 128, 128, 128]
    )
    parser.add_argument("--training-num", type=int, default=10)
    parser.add_argument("--test-num", type=int, default=10)
    parser.add_argument("--logdir", type=str, default="log")
    parser.add_argument("--render", type=float, default=0.1)
    parser.add_argument(
        "--render_mode",
        type=str,
        default="human",
        choices=["human", "rgb_array", "text", "text_full"],
        help="Choose the rendering mode for the game.",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Flag to enable to print extra debugging info",
    )
    parser.add_argument(
        "--self_play",
        action="store_true",
        help="Flag to enable training via self-play (as opposed to fixed opponent)",
    )
    parser.add_argument(
        "--self_play_generations",
        type=int,
        default=5,
        help="Number of generations of self-play agents to train (each generation can have multiple epochs of training)",
    )
    parser.add_argument(
        "--self_play_greedy",
        action="store_true",
        help="Flag to have self-play train against a greedy agent for the first generation",
    )
    parser.add_argument(
        "--cpu-players",
        type=int,
        default=2,
        choices=[1, 2],
        help="Number of CPU players (options: 1, 2)",
    )
    parser.add_argument(
        "--player",
        type=int,
        default=0,
        choices=[0, 1],
        help="Choose which player to play as: red = 0, yellow = 1",
    )
    parser.add_argument(
        "--record",
        action="store_true",
        help="Flag to save a recording of the game (game.gif)",
    )
    parser.add_argument(
        "--win-rate",
        type=float,
        default=0.6,
        help="the expected winning rate: Optimal policy can get 0.7",
    )
    parser.add_argument(
        "--watch",
        default=False,
        action="store_true",
        help="no training, " "watch the play of pre-trained models",
    )
    parser.add_argument(
        "--agent-id",
        type=int,
        default=2,
        help="the learned agent plays as the"
        " agent_id-th player. Choices are 1 and 2.",
    )
    parser.add_argument(
        "--resume-path",
        type=str,
        default="",
        help="the path of agent pth file " "for resuming from a pre-trained agent",
    )
    parser.add_argument(
        "--opponent-path",
        type=str,
        default="",
        help="the path of opponent agent pth file "
        "for resuming from a pre-trained agent",
    )
    parser.add_argument(
        "--device", type=str, default="cuda" if torch.cuda.is_available() else "cpu"
    )
    return parser


def get_args() -> argparse.Namespace:
    parser = get_parser()
    return parser.parse_known_args()[0]


def get_agents(
    args: argparse.Namespace = get_args(),
    agent_learn: Optional[BasePolicy] = None,
    agent_opponent: Optional[BasePolicy] = None,
    optim: Optional[torch.optim.Optimizer] = None,
) -> Tuple[BasePolicy, torch.optim.Optimizer, list]:
    env = get_env()
    observation_space = (
        env.observation_space["observation"]
        if isinstance(env.observation_space, gym.spaces.Dict)
        or isinstance(env.observation_space, gymnasium.spaces.Dict)
        else env.observation_space
    )
    args.state_shape = observation_space.shape or observation_space.n
    args.action_shape = env.action_space.shape or env.action_space.n
    if agent_learn is None:
        # model
        net = Net(
            args.state_shape,
            args.action_shape,
            hidden_sizes=args.hidden_sizes,
            device=args.device,
        ).to(args.device)
        if optim is None:
            optim = torch.optim.Adam(net.parameters(), lr=args.lr)
        agent_learn = DQNPolicy(
            net,
            optim,
            args.gamma,
            args.n_step,
            target_update_freq=args.target_update_freq,
        )
        if args.resume_path:
            agent_learn.load_state_dict(torch.load(args.resume_path))

    if agent_opponent is None:
        if args.self_play:
            # Create a new network with the same shape
            net_opponent = Net(
                args.state_shape,
                args.action_shape,
                hidden_sizes=args.hidden_sizes,
                device=args.device,
            ).to(args.device)
            agent_opponent = DQNPolicy(
                net_opponent,
                optim,
                args.gamma,
                args.n_step,
                target_update_freq=args.target_update_freq,
            )
        elif args.opponent_path:
            agent_opponent = deepcopy(agent_learn)
            agent_opponent.load_state_dict(torch.load(args.opponent_path))
        else:
            # agent_opponent = RandomPolicy()
            agent_opponent = (
                GreedyPolicy()
            )  # Greedy policy is a difficult opponent, should yield much better results than random

    if args.agent_id == 1:
        agents = [agent_learn, agent_opponent]
    else:
        agents = [agent_opponent, agent_learn]
    policy = MultiAgentPolicyManager(agents, env)
    return policy, optim, env.agents


def get_env(render_mode=None, args=None):
    return PettingZooEnv(gobblet_v1.env(render_mode=render_mode, args=args))


def train_selfplay(
    args: argparse.Namespace = get_args(),
    agent_learn: Optional[BasePolicy] = None,
    agent_fixed: Optional[BasePolicy] = None,
):
    # always train first agent, start from random policy

    # ======== environment setup =========
    train_envs = DummyVectorEnv([get_env for _ in range(args.training_num)])
    test_envs = DummyVectorEnv([get_env for _ in range(args.test_num)])
    # seed
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)
    train_envs.seed(args.seed)
    test_envs.seed(args.seed)

    # ======== environment space =========
    # model
    env = get_env()
    observation_space = (
        env.observation_space["observation"]
        if isinstance(env.observation_space, gymnasium.spaces.Dict)
        or isinstance(env.observation_space, gym.spaces.Dict)
        else env.observation_space
    )
    args.state_shape = observation_space.shape or observation_space.n
    args.action_shape = env.action_space.shape or env.action_space.n

    # ======== model setup =========
    # Note: custom models can be specified using the agent_fixed and agent_learn arguments
    if agent_learn is None:
        net = Net(
            args.state_shape,
            args.action_shape,
            hidden_sizes=args.hidden_sizes,
            device=args.device,
        ).to(args.device)
        optim = torch.optim.Adam(net.parameters(), lr=args.lr)
        agent_learn = DQNPolicy(
            net,
            optim,
            args.gamma,
            args.n_step,
            target_update_freq=args.target_update_freq,
        )

    if agent_fixed is None:
        net_fixed = Net(
            args.state_shape,
            args.action_shape,
            hidden_sizes=args.hidden_sizes,
            device=args.device,
        ).to(args.device)
        optim_fixed = torch.optim.SGD(net_fixed.parameters(), lr=0)
        agent_fixed = DQNPolicy(
            net_fixed,
            optim_fixed,
            args.gamma,
            args.n_step,
            target_update_freq=args.target_update_freq,
        )

    # Load initial opponent from file
    # path = os.path.join(args.logdir, 'gobblet_rl', 'dqn', 'policy.pth')
    # agent_fixed.load_state_dict(torch.load(path))

    # ======== agent setup =========
    agents = [agent_learn, agent_fixed]
    policy = MultiAgentPolicyManager(agents, env)
    agents_list = list(policy.policies.keys())

    # ======== collector setup =========
    train_collector = Collector(
        policy,
        train_envs,
        VectorReplayBuffer(args.buffer_size, len(train_envs)),
        exploration_noise=True,
    )
    test_collector = Collector(policy, test_envs, exploration_noise=True)
    # policy.set_eps(1)
    train_collector.collect(n_step=args.batch_size * args.training_num)

    # ======== tensorboard logging setup =========
    log_path = os.path.join(args.logdir, "gobblet_rl", "dqn-selfplay")
    writer = SummaryWriter(log_path)
    writer.add_text("args", str(args))
    logger = TensorboardLogger(writer)

    def save_best_fn(policy):
        if hasattr(args, "model_save_path"):
            model_save_path = args.model_save_path
        else:
            model_save_path = os.path.join(
                args.logdir, "gobblet_rl", "dqn-selfplay", "policy.pth"
            )
        torch.save(
            policy.policies[agents_list[args.agent_id - 1]].state_dict(),
            model_save_path,
        )

    def stop_fn(mean_rewards):
        return mean_rewards >= args.win_rate

    def train_fn(epoch, env_step):
        policy.policies[agents_list[0]].set_eps(args.eps_train)

        if hasattr(policy.policies[agents_list[1]], "set_eps"):
            policy.policies[agents_list[1]].set_eps(args.eps_train)

    def test_fn(epoch, env_step):
        policy.policies[agents_list[0]].set_eps(args.eps_test)

        if hasattr(policy.policies[agents_list[1]], "set_eps"):
            policy.policies[agents_list[1]].set_eps(args.eps_test)

    def reward_metric(rews):
        return rews[:, 0]

    # Self-play loop
    for i in range(args.self_play_generations):
        result = offpolicy_trainer(
            policy,
            train_collector,
            test_collector,
            args.epoch,
            args.step_per_epoch,
            args.step_per_collect,
            args.test_num,
            args.batch_size,
            train_fn=train_fn,
            test_fn=test_fn,
            stop_fn=stop_fn,
            save_best_fn=save_best_fn,
            update_per_step=args.update_per_step,
            logger=logger,
            test_in_train=False,
            reward_metric=reward_metric,
        )

        print(
            f"Launching game between learned agent ({type(policy.policies[agents_list[0]]).__name__}) and fixed agent ({type(policy.policies[agents_list[1]]).__name__}):"
        )
        # Render a single game between the learned policy and fixed policy from last generation
        watch_selfplay(args, policy.policies[agents_list[0]])

        # Set fixed opponent policy as to the current trained policy (updates every epoch)
        policy.policies[agents_list[1]] = deepcopy(policy.policies[agents_list[0]])
        print(f"--- SELF-PLAY GENERATION: {i + 1} ---")

        print(
            f"Launching game between learned agent ({type(policy.policies[agents_list[0]]).__name__}) and itself ({type(policy.policies[agents_list[1]]).__name__}):"
        )
        # Render a single game between the learned policy and itself
        watch_selfplay(args, policy.policies[agents_list[0]])

    model_save_path = os.path.join(
        args.logdir, "gobblet_rl", "dqn-selfplay", "policy.pth"
    )
    torch.save(policy.policies[agents_list[0]].state_dict(), model_save_path)

    return result, policy.policies[agents_list[0]]


def train_agent(
    args: argparse.Namespace = get_args(),
    agent_learn: Optional[BasePolicy] = None,
    agent_opponent: Optional[BasePolicy] = None,
    optim: Optional[torch.optim.Optimizer] = None,
) -> Tuple[dict, BasePolicy]:
    # ======== environment setup =========
    train_envs = DummyVectorEnv([get_env for _ in range(args.training_num)])
    test_envs = DummyVectorEnv([get_env for _ in range(args.test_num)])
    # seed
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)
    train_envs.seed(args.seed)
    test_envs.seed(args.seed)

    # ======== agent setup =========
    policy, optim, agents = get_agents(
        args, agent_learn=agent_learn, agent_opponent=agent_opponent, optim=optim
    )
    agents_list = list(policy.policies.keys())

    # ======== collector setup =========
    train_collector = Collector(
        policy,
        train_envs,
        VectorReplayBuffer(args.buffer_size, len(train_envs)),
        exploration_noise=True,
    )
    test_collector = Collector(policy, test_envs, exploration_noise=True)
    # policy.set_eps(1)
    train_collector.collect(n_step=args.batch_size * args.training_num)

    # ======== tensorboard logging setup =========
    log_path = os.path.join(args.logdir, "gobblet_rl", "dqn")
    writer = SummaryWriter(log_path)
    writer.add_text("args", str(args))
    logger = TensorboardLogger(writer)

    # ======== callback functions used during training =========
    def save_best_fn(policy):
        if hasattr(args, "model_save_path"):
            model_save_path = args.model_save_path
        else:
            model_save_path = os.path.join(
                args.logdir, "gobblet_rl", "dqn", "policy.pth"
            )
        torch.save(
            policy.policies[agents_list[args.agent_id - 1]].state_dict(),
            model_save_path,
        )

    def stop_fn(mean_rewards):
        return mean_rewards >= args.win_rate

    def train_fn(epoch, env_step):
        policy.policies[agents_list[args.agent_id - 1]].set_eps(args.eps_train)

    def test_fn(epoch, env_step):
        policy.policies[agents_list[args.agent_id - 1]].set_eps(args.eps_test)

    def reward_metric(rews):
        return rews[:, args.agent_id - 1]

    # trainer
    result = offpolicy_trainer(
        policy,
        train_collector,
        test_collector,
        args.epoch,
        args.step_per_epoch,
        args.step_per_collect,
        args.test_num,
        args.batch_size,
        train_fn=train_fn,
        test_fn=test_fn,
        stop_fn=stop_fn,
        save_best_fn=save_best_fn,
        update_per_step=args.update_per_step,
        logger=logger,
        test_in_train=False,
        reward_metric=reward_metric,
    )

    return result, policy.policies[agents_list[args.agent_id - 1]]


# ======== a test function that tests a pre-trained agent ======
def watch(
    args: argparse.Namespace = get_args(),
    agent_learn: Optional[BasePolicy] = None,
    agent_opponent: Optional[BasePolicy] = None,
) -> None:
    env = DummyVectorEnv([lambda: get_env(render_mode=args.render_mode, args=args)])
    policy, optim, agents = get_agents(
        args, agent_learn=agent_learn, agent_opponent=agent_opponent
    )
    agents_list = list(policy.policies.keys())

    policy.eval()
    policy.policies[agents_list[args.agent_id - 1]].set_eps(args.eps_test)

    collector = Collector(policy, env, exploration_noise=True)
    pettingzoo_env = env.workers[
        0
    ].env.env  # DummyVectorEnv -> Tianshou PettingZoo Wrapper -> PettingZoo Env
    if args.record:
        recorder = GIFRecorder()
        recorder.capture_frame(pettingzoo_env.unwrapped.screen)
    else:
        recorder = None

    while pettingzoo_env.agents:
        result = collector.collect(n_step=1, render=args.render)
        if recorder is not None:
            recorder.capture_frame(pettingzoo_env.unwrapped.screen)

        time.sleep(0.25)

        if collector.data.terminated or collector.data.truncated:
            rews, lens = result["rews"], result["lens"]
            print(
                f"Final reward: {rews[:, 0].mean()}, length: {lens.mean()} [{type(policy.policies[agents_list[0]]).__name__}]"
            )
            print(
                f"Final reward: {rews[:, 1].mean()}, length: {lens.mean()} [{type(policy.policies[agents_list[1]]).__name__}]"
            )
            if recorder is not None:
                recorder.end_recording(pettingzoo_env.unwrapped.screen)
                recorder = None


def watch_selfplay(
    args: argparse.Namespace = get_args(),
    agent: Optional[BasePolicy] = None,
) -> None:
    # env = get_env(render_mode=args.render_mode, args=args)
    env = DummyVectorEnv([lambda: get_env(render_mode=args.render_mode, args=args)])
    policy = MultiAgentPolicyManager(
        policies=[agent, deepcopy(agent)],
        env=get_env(render_mode=args.render_mode, args=args),
    )  # fixed here
    policy.eval()
    collector = Collector(policy, env, exploration_noise=True)
    result = collector.collect(n_episode=1, render=True)
    rews, lens = result["rews"], result["lens"]
    print(f"Final reward: {rews[:, 0].mean()}, length: {lens.mean()}")


# ======== allows the user to input moves and play vs a pre-trained agent ======
def play(
    args: argparse.Namespace = get_args(),
    agent_learn: Optional[BasePolicy] = None,
    agent_opponent: Optional[BasePolicy] = None,
) -> None:
    env = DummyVectorEnv([lambda: get_env(render_mode=args.render_mode, args=args)])
    # env = get_env(render_mode=args.render_mode, args=args) # Throws error because collector looks for length, could just override though since I'm using my own collector
    policy, optim, agents = get_agents(
        args, agent_learn=agent_learn, agent_opponent=agent_opponent
    )
    agents_list = list(policy.policies.keys())
    policy.eval()
    policy.policies[agents_list[args.agent_id - 1]].set_eps(args.eps_test)

    # Experimental: let the CPU agent to continue training (TODO: check if this actually changes things meaningfully)
    # policy.policies[agents[args.agent_id - 1]].set_eps(args.eps_train)

    collector = ManualPolicyCollector(
        policy, env, exploration_noise=True
    )  # Collector for CPU actions

    pettingzoo_env = env.workers[
        0
    ].env.env  # DummyVectorEnv -> Tianshou PettingZoo Wrapper -> PettingZoo Env

    if args.record:
        recorder = GIFRecorder()
    else:
        recorder = None
    manual_policy = gobblet_v1.ManualGobbletPolicy(
        env=pettingzoo_env, agent_id=args.player, recorder=recorder
    )  # Gobblet keyboard input requires access to raw_env (uses functions from board)
    while pettingzoo_env.agents:
        agent_id = collector.data.obs.agent_id
        # If it is the players turn and there are less than 2 CPU players (at least one human player)
        if agent_id == pettingzoo_env.agents[args.player]:
            observation = {
                "observation": collector.data.obs.obs.flatten(),
                "action_mask": collector.data.obs.mask.flatten(),
            }  # PettingZoo expects a dict with this format
            action = manual_policy(observation, agent_id)

            result = collector.collect_result(
                action=action.reshape(1), render=args.render
            )
        else:
            result = collector.collect(n_step=1, render=args.render)

            if recorder is not None:
                recorder.capture_frame(pettingzoo_env.unwrapped.screen)

        if collector.data.terminated or collector.data.truncated:
            rews, lens = result["rews"], result["lens"]
            print(
                f"Final reward: {rews[:, args.player].mean()}, length: {lens.mean()} [Human]"
            )
            print(
                f"Final reward: {rews[:, 1-args.player].mean()}, length: {lens.mean()} [{type(policy.policies[agents[1-args.player]]).__name__}]"
            )
            if recorder is not None:
                recorder.end_recording(pettingzoo_env.unwrapped.screen)
                recorder = None


if __name__ == "__main__":
    # train the agent and watch its performance in a match!
    args = get_args()
    if args.player == 1:
        args.agent_id = 1  # Ensures trained agent is in the  correct spot

    if args.self_play:
        print("Training agent...")

        # Hard code the first fixed agent to be the greedy policy (after one generation it will be switched to a copy of the learned agent)
        agent_fixed = GreedyPolicy() if args.self_play_greedy else None
        result, agent = train_selfplay(
            args=args, agent_fixed=agent_fixed
        )  # Hard code the first fixed agent to be the greedy policy

        print("Starting game...")
        watch_selfplay(args, agent)
    else:
        print("Training agent...")
        result, agent = train_agent(args)

        print("Starting game...")
        if args.cpu_players == 2:
            watch(args, agent)
        else:
            play(args, agent)



# adapted from https://github.com/Farama-Foundation/PettingZoo/blob/master/tutorials/Tianshou/3_cli_and_logging.py

import argparse
import time
from typing import Tuple

import torch
from tianshou.env import DummyVectorEnv
from tianshou.env.pettingzoo_env import PettingZooEnv
from tianshou.policy import BasePolicy, MultiAgentPolicyManager

from gobblet_rl import gobblet_v1
from gobblet_rl.game.collector_manual_policy import ManualPolicyCollector
from gobblet_rl.game.greedy_policy_tianshou import GreedyPolicy
from gobblet_rl.game.utils import GIFRecorder


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument("--seed", type=int, default=1626)
    parser.add_argument("--eps-test", type=float, default=0.05)
    parser.add_argument("--eps-train", type=float, default=0.1)
    parser.add_argument("--buffer-size", type=int, default=20000)
    parser.add_argument(
        "--lr", type=float, default=1e-4
    )  # TODO: Changing this to 1e-5 for some reason makes it pause after 3 or 4 epochs
    parser.add_argument(
        "--gamma", type=float, default=0.9, help="a smaller gamma favors earlier win"
    )
    parser.add_argument("--n-step", type=int, default=3)
    parser.add_argument("--target-update-freq", type=int, default=320)
    parser.add_argument("--epoch", type=int, default=50)
    parser.add_argument("--step-per-epoch", type=int, default=1000)
    parser.add_argument("--step-per-collect", type=int, default=10)
    parser.add_argument("--update-per-step", type=float, default=0.1)
    parser.add_argument("--batch-size", type=int, default=64)
    parser.add_argument(
        "--hidden-sizes", type=int, nargs="*", default=[128, 128, 128, 128]
    )
    parser.add_argument("--training-num", type=int, default=10)
    parser.add_argument("--test-num", type=int, default=10)
    parser.add_argument("--logdir", type=str, default="log")
    parser.add_argument("--render", type=float, default=0.1)
    parser.add_argument(
        "--render_mode",
        type=str,
        default="human",
        choices=["human", "rgb_array", "text", "text_full"],
        help="Choose the rendering mode for the game.",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Flag to enable to print extra debugging info",
    )
    parser.add_argument(
        "--self_play",
        action="store_true",
        help="Flag to enable training via self-play (as opposed to fixed opponent)",
    )
    parser.add_argument(
        "--cpu-players",
        type=int,
        default=1,
        choices=[1, 2],
        help="Number of CPU players (options: 1, 2)",
    )
    parser.add_argument(
        "--player",
        type=int,
        default=0,
        choices=[0, 1],
        help="Choose which player to play as: red = 0, yellow = 1",
    )
    parser.add_argument(
        "--record",
        action="store_true",
        help="Flag to save a recording of the game (game.gif)",
    )
    parser.add_argument(
        "--depth",
        type=int,
        default=2,
        choices=[1, 2, 3],
        help="Search depth for greedy agent. Options: 1,2,3",
    )
    parser.add_argument(
        "--win-rate",
        type=float,
        default=0.6,
        help="the expected winning rate: Optimal policy can get 0.7",
    )
    parser.add_argument(
        "--watch",
        default=False,
        action="store_true",
        help="no training, " "watch the play of pre-trained models",
    )
    parser.add_argument(
        "--agent-id",
        type=int,
        default=2,
        help="the learned agent plays as the"
        " agent_id-th player. Choices are 1 and 2.",
    )
    parser.add_argument(
        "--resume-path",
        type=str,
        default="",
        help="the path of agent pth file " "for resuming from a pre-trained agent",
    )
    parser.add_argument(
        "--opponent-path",
        type=str,
        default="",
        help="the path of opponent agent pth file "
        "for resuming from a pre-trained agent",
    )
    parser.add_argument(
        "--device", type=str, default="cuda" if torch.cuda.is_available() else "cpu"
    )
    return parser


def get_args() -> argparse.Namespace:
    parser = get_parser()
    return parser.parse_known_args()[0]


def get_agents() -> Tuple[BasePolicy, list]:
    env = get_env()
    agents = [GreedyPolicy(depth=args.depth), GreedyPolicy(depth=args.depth)]
    policy = MultiAgentPolicyManager(agents, env)
    return policy, env.agents


def get_env(render_mode=None, args=None):
    return PettingZooEnv(gobblet_v1.env(render_mode=render_mode, args=args))


# ======== watch two greedy agents play each other (tends to get stuck in loops) ======
def watch() -> None:
    env = DummyVectorEnv([lambda: get_env(render_mode=args.render_mode, args=args)])
    policy, agents = get_agents()

    collector = ManualPolicyCollector(policy, env)

    pettingzoo_env = env.workers[
        0
    ].env.env  # DummyVectorEnv -> Tianshou PettingZoo Wrapper -> PettingZoo Env
    if args.record:
        recorder = GIFRecorder()
    else:
        recorder = None

    while pettingzoo_env.agents:
        result = collector.collect(n_step=1, render=args.render)
        if recorder is not None:
            recorder.capture_frame(pettingzoo_env.unwrapped.screen)
            time.sleep(0.25)

        if collector.data.terminated or collector.data.truncated:
            rews, lens = result["rews"], result["lens"]
            print(
                f"Final reward: {rews[:, 0].mean()}, length: {lens.mean()} [{type(policy.policies[agents[0]]).__name__}]"
            )
            print(
                f"Final reward: {rews[:, 1].mean()}, length: {lens.mean()} [{type(policy.policies[agents[1]]).__name__}]"
            )
            if recorder is not None:
                recorder.end_recording(pettingzoo_env.unwrapped.screen)
                recorder = None


# ======== allows the user to input moves and play vs a greedy agent ======
def play() -> None:
    env = DummyVectorEnv([lambda: get_env(render_mode=args.render_mode, args=args)])

    policy, agents = get_agents()
    collector = ManualPolicyCollector(
        policy, env, exploration_noise=True
    )  # Collector for CPU actions

    pettingzoo_env = env.workers[
        0
    ].env.env  # DummyVectorEnv -> Tianshou PettingZoo Wrapper -> PettingZoo Env
    if args.record:
        recorder = GIFRecorder()
    else:
        recorder = None
    manual_policy = gobblet_v1.ManualGobbletPolicy(
        env=pettingzoo_env, agent_id=args.player, recorder=recorder
    )  # Gobblet keyboard input requires access to raw_env (uses functions from board)

    while pettingzoo_env.agents:
        agent_id = collector.data.obs.agent_id
        # If it is the players turn and there are less than 2 CPU players (at least one human player)
        if agent_id == pettingzoo_env.agents[args.player]:
            observation = {
                "observation": collector.data.obs.obs.flatten(),
                "action_mask": collector.data.obs.mask.flatten(),
            }  # PettingZoo expects a dict with this format
            action = manual_policy(observation, agent_id)

            result = collector.collect_result(
                action=action.reshape(1), render=args.render
            )
        else:
            result = collector.collect(n_step=1, render=args.render)
            if recorder is not None:
                recorder.capture_frame(pettingzoo_env.unwrapped.screen)

        if collector.data.terminated or collector.data.truncated:
            rews, lens = result["rews"], result["lens"]
            print(
                f"\nFinal reward: {rews[:, args.player].mean()}, length: {lens.mean()} [Human]"
            )
            print(
                f"Final reward: {rews[:, 1-args.player].mean()}, length: {lens.mean()} [{type(policy.policies[agents[1-args.player]]).__name__}]"
            )
            if recorder is not None:
                recorder.end_recording(pettingzoo_env.unwrapped.screen)


if __name__ == "__main__":
    # train the agent and watch its performance in a match!
    args = get_args()
    print("Starting game...")
    if args.cpu_players == 2:
        watch()
    else:
        play()



# adapted from https://github.com/ray-project/ray/blob/master/rllib/examples/self_play_with_open_spiel.py
"""Example showing how one can implement a simple self-play training workflow using PettingZoo.

Uses the PettingZoo adapter of RLlib with the "gobblet" game and
a multi-agent setup with a "main" policy and n "main_v[x]" policies
(x=version number), which are all at-some-point-frozen copies of
"main". At the very beginning, "main" plays against RandomPolicy.

Checks for the training progress after each training update via a custom
callback. We simply measure the win rate of "main" vs the opponent
("main_v[x]" or RandomPolicy at the beginning) by looking through the
achieved rewards in the episodes in the train batch. If this win rate
reaches some configurable threshold, we add a new policy to
the policy map (a frozen copy of the current "main" one) and change the
policy_mapping_fn to make new matches of "main" vs any of the previous
versions of "main" (including the just added one).

After training for n iterations, a configurable number of episodes can
be played by the user against the "main" agent on the command line.
"""

import argparse
import os
import sys

import gymnasium
import numpy as np
import ray
import supersuit as ss
from ray import air, tune
from ray.rllib.algorithms.callbacks import DefaultCallbacks
from ray.rllib.algorithms.ppo import PPOConfig
from ray.rllib.env.wrappers.pettingzoo_env import PettingZooEnv
from ray.rllib.examples.policy.random_policy import RandomPolicy
from ray.rllib.policy.policy import PolicySpec
from ray.tune import CLIReporter, register_env

from gobblet_rl import gobblet_v1
from gobblet_rl.game.greedy_policy_rllib import GreedyPolicy
from gobblet_rl.game.random_admissible_policy_rllib import RandomAdmissiblePolicy

parser = argparse.ArgumentParser()
parser.add_argument(
    "--framework",
    choices=["tf", "tf2", "torch"],
    default="tf",
    help="The DL framework specifier.",
)
parser.add_argument("--num-cpus", type=int, default=0)
parser.add_argument("--num-workers", type=int, default=2)
parser.add_argument(
    "--from-checkpoint",
    type=str,
    default=None,
    help="Full path to a checkpoint file for restoring a previously saved "
    "Algorithm state.",
)
parser.add_argument(
    "--stop-iters", type=int, default=200, help="Number of iterations to train."
)
parser.add_argument(
    "--stop-timesteps", type=int, default=10000000, help="Number of timesteps to train."
)
parser.add_argument(
    "--seed", type=int, default=0, help="Seed for random number generation."
)
parser.add_argument(
    "--win-rate-threshold",
    type=float,
    default=0.95,
    help="Win-rate at which we setup another opponent by freezing the "
    "current main policy and playing against a uniform distribution "
    "of previously frozen 'main's from here on.",
)
parser.add_argument(
    "--num-episodes-human-play",
    type=int,
    default=10,
    help="How many episodes to play against the user on the command "
    "line after training has finished.",
)
parser.add_argument(
    "--render-mode",
    type=str,
    default="human",
    choices=["human", "text", "text_full", "rgb_array"],
    help="Choose rendering mode (options: human, text, text_full, rgb_array)",
)
parser.add_argument(
    "--opponent-policy",
    type=str,
    default="random_admissible",
    choices=["random", "random_admissible", "greedy"],
    help="Choose which policy the agent plays against in the first iteration (choices: random, random_admissible, greedy)",
)
parser.add_argument(
    "--no-masking",
    action="store_true",
)
args = parser.parse_args()


def ask_user_for_action(observation):
    """Asks the user for a valid action on the command line and returns it.

    Re-queries the user until she picks a valid one.

    Args:
        observation: The observation.
    """
    action_mask = observation["action_mask"]
    legal_moves = action_mask.nonzero()[0]
    choice = -1
    while choice not in legal_moves:
        print(f"Choose an action from {legal_moves}:")
        sys.stdout.flush()
        choice_str = input()
        try:
            choice = int(choice_str)
        except ValueError:
            continue
    return choice


class SelfPlayCallback(DefaultCallbacks):
    def __init__(self):
        super().__init__()
        # 0=RandomPolicy, 1=1st main policy snapshot,
        # 2=2nd main policy snapshot, etc..
        self.current_opponent = 0

    def on_train_result(self, *, algorithm, result, **kwargs):
        # Get the win rate for the train batch.
        # Note that normally, one should set up a proper evaluation config,
        # such that evaluation always happens on the already updated policy,
        # instead of on the already used train_batch.
        main_rew = result["hist_stats"].pop(
            "episode_reward"
        )  # Otherwise, key error "policy_main_reward"
        opponent_rew = list(result["hist_stats"].values())[0]
        assert len(main_rew) == len(opponent_rew)
        won = 0
        for r_main, r_opponent in zip(main_rew, opponent_rew):
            if r_main > r_opponent:
                won += 1
        win_rate = won / len(main_rew)
        result["win_rate"] = win_rate
        print(f"Iter={algorithm.iteration} win-rate={win_rate} -> ", end="")
        # If win rate is good -> Snapshot current policy and play against
        # it next, keeping the snapshot fixed and only improving the "main"
        # policy.
        if win_rate > args.win_rate_threshold:
            self.current_opponent += 1
            new_pol_id = f"main_v{self.current_opponent}"
            print(f"adding new opponent to the mix ({new_pol_id}).")

            # Re-define the mapping function, such that "main" is forced
            # to play against any of the previously played policies
            # (excluding "random").
            def policy_mapping_fn(agent_id, episode, worker, **kwargs):
                # agent_id = [0|1] -> policy depends on episode ID
                # This way, we make sure that both policies sometimes play
                # (start player) and sometimes agent1 (player to move 2nd).
                return (
                    "main"
                    if episode.episode_id % 2 == agent_id
                    else "main_v{}".format(
                        np.random.choice(list(range(1, self.current_opponent + 1)))
                    )
                )

            new_policy = algorithm.add_policy(
                policy_id=new_pol_id,
                policy_cls=type(algorithm.get_policy("main")),
                policy_mapping_fn=policy_mapping_fn,
            )

            # Set the weights of the new policy to the main policy.
            # We'll keep training the main policy, whereas `new_pol_id` will
            # remain fixed.
            main_state = algorithm.get_policy("main").get_state()
            new_policy.set_state(main_state)
            # We need to sync the just copied local weights (from main policy)
            # to all the remote workers as well.
            algorithm.workers.sync_weights()
        else:
            print("not good enough; will keep learning ...")

        # +2 = main + random
        result["league_size"] = self.current_opponent + 2


if __name__ == "__main__":
    ray.init(num_cpus=args.num_cpus or None, include_dashboard=False, local_mode=True)

    OPPONENT_POLICIES = {
        "random": RandomPolicy,
        "random_admissible": RandomAdmissiblePolicy,
        "greedy": GreedyPolicy,
    }
    OpponentPolicy = OPPONENT_POLICIES[args.opponent_policy]

    def convert_box(convert_obs_fn, old_box):
        new_low = convert_obs_fn(old_box.low)
        new_high = convert_obs_fn(old_box.high)
        return gymnasium.spaces.Box(low=new_low, high=new_high, dtype=new_low.dtype)

    # Changes observation space from (6,7,2) to (84,)
    def flattened_pettingzoo_env(render_mode=None):
        env = gobblet_v1.env(render_mode=render_mode)
        shape = env.observation_space(env.possible_agents[0])["observation"].shape
        newshape = np.product(shape).reshape(1)

        def change_obs_space_fn(obs_space):
            obs_space["observation"] = convert_box(
                lambda obs: obs.reshape(newshape), old_box=obs_space["observation"]
            )
            return obs_space

        def change_observation_fn(observation, old_obs_space):
            observation["observation"] = observation["observation"].reshape(newshape)
            return observation

        env = ss.lambda_wrappers.observation_lambda_v0(
            env,
            change_obs_space_fn=change_obs_space_fn,
            change_observation_fn=change_observation_fn,
        )
        return env

    def env_creator(args):
        # return ParallelPettingZooEnv(flattened_pettingzoo_env())
        return PettingZooEnv(flattened_pettingzoo_env())

    env = env_creator({})
    register_env("gobblet_rl", env_creator)

    def policy_mapping_fn(agent_id, episode, worker, **kwargs):
        # agent_id = [0|1] -> policy depends on episode ID
        # This way, we make sure that both policies sometimes play agent0
        # (start player) and sometimes agent1 (player to move 2nd).
        return "main" if episode.episode_id % 2 == agent_id else "random"

    config = (
        PPOConfig()
        .environment("gobblet_rl")
        .framework(args.framework)
        .callbacks(SelfPlayCallback)
        .rollouts(num_envs_per_worker=5, num_rollout_workers=args.num_workers)
        .training(
            # num_sgd_iter=20,
            # model={"fcnet_hiddens": [512, 512]}
            # the ActionMaskModel retrieves the invalid actions and avoids them
            # model={
            #     "custom_model": ActionMaskModel
            #     if args.framework != "torch"
            #     else TorchActionMaskModel,
            #     # disable action masking according to CLI
            #     "custom_model_config": {"no_masking": args.no_masking},
            # },
        )
        .multi_agent(
            # Initial policy map: Random and PPO. This will be expanded
            # to more policy snapshots taken from "main" against which "main"
            # will then play (instead of "random"). This is done in the
            # custom callback defined above (`SelfPlayCallback`).
            policies={
                # Our main policy, we'd like to optimize.
                "main": PolicySpec(
                    observation_space=env.observation_space,
                    action_space=env.action_space,
                ),
                # An initial random opponent to play against.
                "random": PolicySpec(
                    observation_space=env.observation_space,
                    action_space=env.action_space,
                    policy_class=OpponentPolicy,
                ),
            },
            # Assign agent 0 and 1 randomly to the "main" policy or
            # to the opponent ("random" at first). Make sure (via episode_id)
            # that "main" always plays against "random" (and not against
            # another "main").
            policy_mapping_fn=policy_mapping_fn,
            # Always just train the "main" policy.
            policies_to_train=["main"],
        )
        # Use GPUs iff `RLLIB_NUM_GPUS` env var set to > 0.
        .resources(num_gpus=int(os.environ.get("RLLIB_NUM_GPUS", "0")))
    )
    config.preprocessor_pref = None
    config._disable_preprocessor_api = True

    stop = {
        "timesteps_total": args.stop_timesteps,
        "training_iteration": args.stop_iters,
    }

    # Train the "main" policy to play really well using self-play.
    results = None
    if not args.from_checkpoint:
        results = tune.Tuner(
            "PPO",
            param_space=config.to_dict(),
            run_config=air.RunConfig(
                stop=stop,
                verbose=2,
                progress_reporter=CLIReporter(
                    metric_columns={
                        "training_iteration": "iter",
                        "time_total_s": "time_total_s",
                        "timesteps_total": "ts",
                        "episodes_this_iter": "train_episodes",
                        "policy_reward_mean/main": "reward",
                        "win_rate": "win_rate",
                        "league_size": "league_size",
                    },
                    sort_by_metric=True,
                ),
                checkpoint_config=air.CheckpointConfig(
                    checkpoint_at_end=True,
                    checkpoint_frequency=10,
                ),
            ),
        ).fit()

    # Restore trained trainer (set to non-explore behavior) and play against
    # human on command line.
    if args.num_episodes_human_play > 0:
        num_episodes = 0
        config.explore = False
        algo = config.build()
        if args.from_checkpoint:
            algo.restore(args.from_checkpoint)
        else:
            checkpoint = results.get_best_result().checkpoint
            if not checkpoint:
                raise ValueError("No last checkpoint found in results!")
            algo.restore(checkpoint)

        # Play from the against the trained agent
        # in an actual (non-RLlib-wrapped) PettingZoo env.
        human_player = 1

        env = flattened_pettingzoo_env(render_mode=args.render_mode)

        while num_episodes < args.num_episodes_human_play:
            print("\nYou play as {}".format("yellow" if human_player else "red"))
            env.reset()
            env.render()

            manual_policy = gobblet_v1.ManualGobbletPolicy(
                env=env, agent_id=human_player
            )  # Has to be called after env.reset as it accesses the env's agents

            for agent in env.agent_iter():
                observation, reward, termination, truncation, info = env.last()
                if termination or truncation:
                    agent_name = (
                        "human"
                        if agent == env.possible_agents[human_player]
                        else f"CPU ({algo.config.algo_class.__name__})"
                    )
                    print(f"Agent: {agent_name}, Reward: {reward}")
                    if agent_name == "human":
                        if reward > 0:
                            print("You win")
                        elif reward < 0:
                            print("You lose")
                        else:
                            print("Draw")
                    env.step(None)
                else:
                    if agent == env.possible_agents[human_player]:
                        if args.render_mode == "human":
                            action = manual_policy(
                                observation, agent
                            )  # Get user input from Pygame
                        else:
                            action = ask_user_for_action(
                                observation
                            )  # Get user input from command line
                    else:
                        action = algo.compute_single_action(
                            observation, policy_id="main"
                        )
                        # In case computer chooses an invalid action, pick a
                        # random one.
                        action_mask = observation["action_mask"]
                        if action_mask[action] == 0:
                            action = np.random.choice(
                                np.arange(len(action_mask)),
                                p=action_mask / np.sum(action_mask),
                            )
                    env.step(action)
                env.render()
            env.render()

            print("End of game!")

            # Switch order of players
            human_player = 1 - human_player

            num_episodes += 1

        env.render()  # TODO do we need this
        algo.stop()

    ray.shutdown()



# noqa
"""
# Gobblet

```{figure} classic_tictactoe.gif # TODO: gif of gobblet?
:width: 140px
:name: Gobblet
```

This environment is part of the <a href='..'>classic environments</a>. Please read that page first for general information.

| Import             | `from gobblet_rl import gobblet_v1`   |
|--------------------|------------------------------------|
| Actions            | Discrete                           |
| Parallel API       | Yes                                |
| Manual Control     | No                                 |
| Agents             | `agents= ['player_1', 'player_2']` |
| Agents             | 2                                  |
| Action Shape       | Discrete(53)                       |
| Action Values      | Discrete(53)                       |
| Observation Shape  | (3,3,15)                           |
| Observation Values | [0,1]                              |


Gobblet is a simple turn based strategy game based on tic-tac-toe, where 2 players, white and black, take turns placing pieces on a 3 x 3 grid.
They can choose from three sizes of pieces: small, medium and large. Each player has two pieces of each size, making six total.
Players can either place a piece in an empty space, "gobble" an opponent's piece with a larger piece, or move an already placed piece either to an empty space or to gobble an opponent's piece.
The first player to place 3 of their pieces in a horizontal, vertical, or diagonal line is the winner.

### Observation Space

The observation is a dictionary which contains an `'observation'` element which is the usual RL observation described below, and an  `'action_mask'` which holds the legal moves, described in the Legal Actions Mask section.

Inspired by AlphaZero's observation space for Chess, the main observation space is a 3x3 image representing the board. It has channels representing:

* Channel 0: Is black or white
* Channels 1 - 12: One channel for each piece type and player combination. For example, channels 1-2 represent white small pieces, 3-4 white medium, and 5-6 white large pieces.
An index of this channel is set to 1 if a white piece is in the corresponding spot on the game board, otherwise it is set to 0.
* Channel 13: A move clock counting up to the 50 move rule. Represented by a single channel where the *n* th element in the flattened channel is set if there has been *n* moves
* Channel 14: All ones to help neural networks find board edges in padded convolutions
* Channel 15: represents whether a position has been seen before (whether a position is a 2-fold repetition)
# TODO: do we need channels 0 and 13-15? Testing without at first

Like AlphaZero, the board is always oriented towards the current agent (the currant agent's king starts on the 1st row). In other words, the two players are looking at mirror images of the board, not the same board.

Unlike AlphaZero, the observation space does not stack the observations previous moves by default. This can be accomplished using the `frame_stacking` argument of our wrapper.

#### Legal Actions Mask

The legal moves available to the current agent are found in the `action_mask` element of the dictionary observation. The `action_mask` is a binary vector where each index of the vector represents whether the action is legal or not. The `action_mask` will be all zeros for any agent except the one
whose turn it is. Taking an illegal move ends the game with a reward of -1 for the illegally moving agent and a reward of 0 for all other agents.

### Action Space

Actions range from 0 - 53, representing possible positions on the board to place each piece. We count moving already placed pieces the same as placing them regularly.
Actions 0-8 represent positions on the board to place the first small piece.
Actions 9-17 represent positions on the board to place the second small piece.
Actions 18-26 represent positions on the board to place the first medium piece.
Actions 27-35 represent positions on the board to place the second medium piece.
Actions 36-44 represent positions on the board to place the first large piece.
Actions 45-53 represent positions on the board to place the second large piece.

The cells in the board are indexed as follows:

 ```
0 | 3 | 6
_________

1 | 4 | 7
_________

2 | 5 | 8
 ```

To determine the position from an action, we take the number modulo 9, resulting in a number 0-8
To determine the piece from an action i, we use floor division by 9 (i // 9), resulting in a number 0-5, where 1-2 represent small pieces, 3-4 represent medium pieces, and 5-6 represent large pieces.
To determine the size of a given piece p (1-6), we use floor division by 2 (p + 1 // 2) resulting in a number 1-3

### Rewards

### Rewards

| Winner | Loser |
| :----: | :---: |
| +1     | -1    |

If the game ends in a draw, both players will receive a reward of 0.

### Version History

* v1: Added pygame rendering and human interaction
* v0: Initial versions release (1.0.0)

"""

import os

import gymnasium
import numpy as np
import pygame
from gymnasium import spaces
from pettingzoo import AECEnv
from pettingzoo.utils import agent_selector, wrappers
from pettingzoo.utils.conversions import parallel_wrapper_fn

from .board import Board
from .utils import get_image, load_chip, load_chip_preview


def env(render_mode=None, args=None):
    env = raw_env(render_mode=render_mode, args=args)
    if render_mode == "ansi":
        env = wrappers.CaptureStdoutWrapper(env)
    env = wrappers.TerminateIllegalWrapper(env, illegal_reward=-1)
    env = wrappers.AssertOutOfBoundsWrapper(env)
    env = wrappers.OrderEnforcingWrapper(env)
    return env


parallel_env = parallel_wrapper_fn(env)


class raw_env(AECEnv):
    metadata = {
        "render_modes": ["human", "rgb_array", "text", "text_full"],
        "name": "gobblet_v1",
        "is_parallelizable": True,
        "render_fps": 60,
        "has_manual_policy": True,
    }

    def __init__(self, render_mode=None, args=None):
        super().__init__()
        self.board = Board()
        self.board_size = 3  # Will need to make a separate file for 4x4

        self.agents = ["player_1", "player_2"]
        self.possible_agents = self.agents[:]

        self.action_spaces = {i: spaces.Discrete(54) for i in self.agents}
        self.observation_spaces = {
            i: spaces.Dict(
                {
                    "observation": spaces.Box(
                        low=0, high=1, shape=(3, 3, 13), dtype=np.int8
                    ),
                    "action_mask": spaces.Box(
                        low=0, high=1, shape=(54,), dtype=np.int8
                    ),
                }
            )
            for i in self.agents
        }

        self.rewards = {i: 0 for i in self.agents}
        self.terminations = {i: False for i in self.agents}
        self.truncations = {i: False for i in self.agents}
        self.infos = {i: {"legal_moves": list(range(0, 9))} for i in self.agents}

        self._agent_selector = agent_selector(self.agents)
        self.agent_selection = self._agent_selector.reset()

        self.render_mode = render_mode
        self.debug = args.debug if hasattr(args, "debug") else False
        self.screen_width = args.screen_width if hasattr(args, "screen_width") else 640
        self.screen_height = self.screen_width
        self.screen = None

    # Key
    # ----
    # blank space = 0
    # agent 0 = 1
    # agent 1 = 2
    # An observation is list of lists, where each list represents a row
    #
    # [[0,0,2]
    #  [1,2,1]
    #  [2,1,0]]
    def observe(self, agent):
        board = self.board.squares.reshape(3, 3, 3)

        if self.agents.index(agent) == 1:
            board = (
                board * -1
            )  # Swap the signs on the board for the two different agents

        # Represent observations in the same way as pettingzoo.chess: specific channel for each color piece (e.g., two for each white small piece)
        layers = []
        for i in range(1, 7):
            layers.append(
                board[(i - 1) // 2] == i
            )  # 3x3 array with an entry of 1 for squares with each white piece (1, ..., 6)

        for i in range(1, 7):
            layers.append(
                board[(i - 1) // 2] == -i
            )  # 3x3 array with an entry of 1 for squares with each black piece (-1, ..., -6)

        if self.agents.index(agent) == 1:
            agents_layer = np.ones((3, 3))
        else:
            agents_layer = np.zeros((3, 3))

        layers.append(
            agents_layer
        )  # Thirteenth layer encoding the current player (agent index 0 or 1)

        observation = np.stack(layers, axis=2).astype(np.int8)
        legal_moves = self._legal_moves() if agent == self.agent_selection else []

        action_mask = np.zeros(54, "int8")
        for i in legal_moves:
            action_mask[i] = 1

        return {"observation": observation, "action_mask": action_mask}

    def observation_space(self, agent):
        return self.observation_spaces[agent]

    def action_space(self, agent):
        return self.action_spaces[agent]

    def _legal_moves(self):
        legal_moves = []
        for action in range(54):
            if self.board.is_legal(action, self.agents.index(self.agent_selection)):
                legal_moves.append(action)
        return legal_moves

    # is a value from 0 to 8 indicating position to move on 3x3 board
    def step(self, action):
        if (
            self.terminations[self.agent_selection]
            or self.truncations[self.agent_selection]
        ):
            return self._was_dead_step(action)
        # check if input action is a valid move (0 == empty spot)
        if not self.board.is_legal(action, self.agent_selection) and self.debug:
            print("piece: ", self.board.get_piece_from_action(action))
            print("piece_size: ", self.board.get_piece_size_from_action(action))
            print("pos: ", self.board.get_pos_from_action(action))
            print("--ERROR-- ILLEGAL MOVE")

        self.board.play_turn(self.agents.index(self.agent_selection), action)

        next_agent = self._agent_selector.next()

        if self.board.check_game_over():
            winner = self.board.check_for_winner()
            if winner == 0:  # NOTE: don't think ties are possible in gobblet
                # tie
                pass
            elif winner == 1:
                # agent 0 won
                self.rewards[self.agents[0]] += 1
                self.rewards[self.agents[1]] -= 1
            else:
                # agent 1 won
                self.rewards[self.agents[1]] += 1
                self.rewards[self.agents[0]] -= 1

            # once either play wins or there is a draw, game over, both players are done
            self.terminations = {i: True for i in self.agents}

        # Switch selection to next agents
        self._cumulative_rewards[self.agent_selection] = 0
        self.agent_selection = next_agent

        self._accumulate_rewards()
        self.turn += 1
        self.action = action
        if self.render_mode in ["human", "text", "text_full", "rgb_array"]:
            self.render()

    def reset(self, seed=None, return_info=False, options=None):
        # reset environment
        self.board = Board()

        self.agents = self.possible_agents[:]
        self.rewards = {i: 0 for i in self.agents}
        self._cumulative_rewards = {i: 0 for i in self.agents}
        self.terminations = {i: False for i in self.agents}
        self.truncations = {i: False for i in self.agents}
        self.infos = {i: {} for i in self.agents}
        # selects the first agent
        self._agent_selector.reinit(self.agents)
        self._agent_selector.reset()
        self.agent_selection = self._agent_selector.reset()
        self.turn = 0
        self.action = -1

    def render(self):
        if self.render_mode is None:
            gymnasium.logger.warn(
                "You are calling render method without specifying any render mode."
            )
            return

        def getSymbolFull(input):
            if input == 0:
                return "- "
            if input > 0:
                return f"+{int(input)}"
            else:
                return f"{int(input)}"

        def getSymbol(input):
            if input == 0:
                return "- "
            if input > 0:
                return f"+{int((input + 1) // 2)}"
            else:
                return f"{int((input) // 2)}"

        if self.debug:
            self.board.print_pieces()
        if self.render_mode == "text" or self.debug:
            pos = self.action % 9
            piece = (self.action // 9) + 1
            piece = (piece + 1) // 2
            print(
                f"TURN: {self.turn}, AGENT: {self.agent_selection}, ACTION: {self.action}, POSITION: {pos}, PIECE: {piece}"
            )
            board = list(map(getSymbol, self.board.get_flatboard()))
            print(" " * 7 + "|" + " " * 7 + "|" + " " * 7)
            print(
                f"  {board[0]}   " + "|" + f"   {board[3]}  " + "|" + f"   {board[6]}  "
            )
            print("_" * 7 + "|" + "_" * 7 + "|" + "_" * 7)

            print(" " * 7 + "|" + " " * 7 + "|" + " " * 7)
            print(
                f"  {board[1]}   " + "|" + f"   {board[4]}  " + "|" + f"   {board[7]}  "
            )
            print("_" * 7 + "|" + "_" * 7 + "|" + "_" * 7)

            print(" " * 7 + "|" + " " * 7 + "|" + " " * 7)
            print(
                f"  {board[2]}   " + "|" + f"   {board[5]}  " + "|" + f"   {board[8]}  "
            )
            print(" " * 7 + "|" + " " * 7 + "|" + " " * 7)
            print()

        elif self.render_mode == "text_full":
            pos = self.action % 9
            piece = (self.action // 9) + 1
            print(
                f"TURN: {self.turn}, AGENT: {self.agent_selection}, ACTION: {self.action}, POSITION: {pos}, PIECE: {piece}"
            )
            board = list(map(getSymbolFull, self.board.squares))
            print(
                " " * 9
                + "SMALL"
                + " " * 9
                + "  "
                + " " * 10
                + "MED"
                + " " * 10
                + "  "
                + " " * 9
                + "LARGE"
                + " " * 9
                + "  "
            )
            top = " " * 7 + "|" + " " * 7 + "|" + " " * 7
            bottom = "_" * 7 + "|" + "_" * 7 + "|" + "_" * 7
            top1 = (
                f"  {board[0]}   " + "|" + f"   {board[3]}  " + "|" + f"   {board[6]}  "
            )
            top2 = (
                f"  {board[9]}   "
                + "|"
                + f"   {board[12]}  "
                + "|"
                + f"   {board[15]}  "
            )
            top3 = (
                f"  {board[18]}   "
                + "|"
                + f"   {board[21]}  "
                + "|"
                + f"   {board[24]}  "
            )
            print(top + "  " + top + "  " + top)
            print(top1 + "  " + top2 + "  " + top3)
            print(bottom + "  " + bottom + "  " + bottom)

            mid1 = (
                f"  {board[1]}   " + "|" + f"   {board[4]}  " + "|" + f"   {board[7]}  "
            )
            mid2 = (
                f"  {board[10]}   "
                + "|"
                + f"   {board[13]}  "
                + "|"
                + f"   {board[16]}  "
            )
            mid3 = (
                f"  {board[19]}   "
                + "|"
                + f"   {board[22]}  "
                + "|"
                + f"   {board[25]}  "
            )
            print(top + "  " + top + "  " + top)
            print(mid1 + "  " + mid2 + "  " + mid3)
            print(bottom + "  " + bottom + "  " + bottom)

            bot1 = (
                f"  {board[2]}   " + "|" + f"   {board[5]}  " + "|" + f"   {board[8]}  "
            )
            bot2 = (
                f"  {board[9+2]}   "
                + "|"
                + f"   {board[9+5]}  "
                + "|"
                + f"   {board[9+8]}  "
            )
            bot3 = (
                f"  {board[18+2]}   "
                + "|"
                + f"   {board[18+5]}  "
                + "|"
                + f"   {board[18+8]}  "
            )
            print(top + "  " + top + "  " + top)
            print(bot1 + "  " + bot2 + "  " + bot3)
            print(top + "  " + top + "  " + top)
            print()

        else:
            # Adapted screen logic from PettingZoo connect_four.py
            if self.render_mode == "human":
                if self.screen is None:
                    pygame.init()
                    self.screen = pygame.display.set_mode(
                        (self.screen_width, self.screen_height)
                    )
                pygame.event.get()
            elif self.screen is None:
                self.screen = pygame.Surface((self.screen_width, self.screen_height))

            # Load and scale all the necessary images
            # Settled on these numbers by trial & error (happens to make things fit nicely)
            tile_size = (self.screen_width * ((47 - 7) / 47)) / 3
            scale_large = 9 / 13
            scale_med = 6 / 13
            scale_small = 4 / 13
            red = {}
            red[3] = load_chip(tile_size, "GobbletLargeRed.png", scale_large)
            red[2] = load_chip(tile_size, "GobbletMedRed.png", scale_med)
            red[1] = load_chip(tile_size, "GobbletSmallRed.png", scale_small)

            yellow = {}
            yellow[3] = load_chip(tile_size, "GobbletLargeYellow.png", scale_large)
            yellow[2] = load_chip(tile_size, "GobbletMedYellow.png", scale_med)
            yellow[1] = load_chip(tile_size, "GobbletSmallYellow.png", scale_small)

            self.preview = {}
            self.preview["player_1"] = {}
            self.preview["player_1"][3] = load_chip_preview(
                tile_size, "GobbletLargeRedPreview.png", scale_large
            )
            self.preview["player_1"][2] = load_chip_preview(
                tile_size, "GobbletMedRedPreview.png", scale_med
            )
            self.preview["player_1"][1] = load_chip_preview(
                tile_size, "GobbletSmallRedPreview.png", scale_small
            )

            self.preview["player_2"] = {}
            self.preview["player_2"][3] = load_chip_preview(
                tile_size, "GobbletLargeYellowPreview.png", scale_large
            )
            self.preview["player_2"][2] = load_chip_preview(
                tile_size, "GobbletMedYellowPreview.png", scale_med
            )
            self.preview["player_2"][1] = load_chip_preview(
                tile_size, "GobbletSmallYellowPreview.png", scale_small
            )

            # preview_chips = {self.agents[0]: self.preview["player_1"], self.agents[1]: self.preview["player_1"]}

            board_img = get_image(os.path.join("img", "GobbletBoard3x3.png"))
            board_img = pygame.transform.scale(
                board_img, ((int(self.screen_width)), int(self.screen_height))
            )

            self.screen.blit(board_img, (0, 0))

            offset = self.screen_width * (
                (9 + 4) / 47
            )  # Piece is 9px wide, gap between pieces 4px, total width is 47px
            offset_side = (
                self.screen_width * (6 / 47)
            ) - 1  # Distance from the side of the board to the first piece is 6px
            offset_centering = offset * 1 / 3 + (
                5 * self.screen_width / 1000
                if self.screen_width > 500
                else 8 * self.screen_width / 1000
            )
            # Extra 5px fixed alignment issues at 1000x1000, but ratio needs to be higher for lower res (trial & error)

            # Blit the chips and their positions
            for i in range(9):
                for j in range(1, 4):
                    if (
                        self.board.squares[i + 9 * (j - 1)] == 2 * j - 1
                        or self.board.squares[i + 9 * (j - 1)] == 2 * j
                    ):  # small pieces (1,2), medium pieces (3,4), large pieces (5,6)
                        self.screen.blit(
                            red[j],
                            red[j].get_rect(
                                center=(
                                    int(i / 3) * (offset)
                                    + offset_side
                                    + offset_centering,
                                    (i % 3) * (offset) + offset_side + offset_centering,
                                )
                            ),
                        )
                    if self.board.squares[i + 9 * (j - 1)] == -1 * (
                        2 * j - 1
                    ) or self.board.squares[i + 9 * (j - 1)] == -1 * (2 * j):
                        self.screen.blit(
                            yellow[j],
                            yellow[j].get_rect(
                                center=(
                                    int(i / 3) * (offset)
                                    + offset_side
                                    + offset_centering,
                                    (i % 3) * (offset) + offset_side + offset_centering,
                                )
                            ),
                        )

            # Blit the preview chips and their positions
            for i in range(9):
                for j in range(1, 4):
                    if self.board.squares_preview[i + 9 * (j - 1)] == 1:
                        self.screen.blit(
                            self.preview["player_1"][j],
                            self.preview["player_1"][j].get_rect(
                                center=(
                                    int(i / 3) * (offset)
                                    + offset_side
                                    + offset_centering,
                                    (i % 3) * (offset) + offset_side + offset_centering,
                                )
                            ),
                        )
                    if self.board.squares_preview[i + 9 * (j - 1)] == -1:
                        self.screen.blit(
                            self.preview["player_2"][j],
                            self.preview["player_2"][j].get_rect(
                                center=(
                                    int(i / 3) * (offset)
                                    + offset_side
                                    + offset_centering,
                                    (i % 3) * (offset) + offset_side + offset_centering,
                                )
                            ),
                        )

            if self.render_mode == "human":
                pygame.display.update()
            observation = np.array(pygame.surfarray.pixels3d(self.screen))

        return (
            np.transpose(observation, axes=(1, 0, 2))
            if self.render_mode == "rgb_array"
            else None
        )

    def close(self):
        if self.screen is not None:
            import pygame

            pygame.quit()
            self.screen = None



import glob
import os
from typing import Tuple

import ray.tune
from ray import init
from ray.rllib.agents import ppo
from ray.rllib.env import PettingZooEnv
from ray.rllib.models import ModelCatalog
from ray.rllib.utils.framework import try_import_torch
from ray.tune.logger import pretty_print
from ray.tune.registry import register_env

from gobblet_rl import gobblet_v1
from gobblet_rl.game.utils import get_project_root
from gobblet_rl.models.action_mask_model import TorchActionMaskModel

torch, nn = try_import_torch()


def prepare_train() -> Tuple[ppo.PPOTrainer, PettingZooEnv]:
    env_name = "pettingzoo_gobblet"

    # get the Pettingzoo env
    def env_creator():
        env = gobblet_v1.env(render_mode=None)
        return env

    register_env(env_name, lambda config: PettingZooEnv(env_creator()))
    ModelCatalog.register_custom_model("pa_model2", TorchActionMaskModel)
    # wrap the pettingzoo env in MultiAgent RLLib
    env = PettingZooEnv(env_creator())

    agents = ["player_1", "player_2"]
    custom_config = {
        "env": env_name,
        "model": {
            "custom_model": "pa_model2",
        },
        "framework": "torch",
        # Use GPUs iff `RLLIB_NUM_GPUS` env var set to > 0.
        # "num_gpus": int(torch.cuda.device_count()),
        "num_gpus": 0,
        "num_workers": 2,  # os.cpu_count() - 1,
        "multiagent": {
            "policies": {
                name: (None, env.observation_space, env.action_space, {})
                for name in agents
            },
            "policy_mapping_fn": lambda agent_id: agent_id,
        },
    }

    # get trainer

    ppo_config = ppo.DEFAULT_CONFIG.copy()
    ppo_config.update(custom_config)

    trainer = ppo.PPOTrainer(config=ppo_config)

    return trainer, env, ppo_config


def train(trainer, max_steps=2e6):
    # run manual training loop and print results after each iteration
    iters = 0
    while True:
        iters += 1
        result = trainer.train()
        print(pretty_print(result))
        # stop training if the target train steps or reward are reached
        if result["timesteps_total"] >= max_steps:
            print(
                "training done, because max_steps"
                f"{max_steps} {result['timesteps_total']} reached"
            )
            break
    # manual test loop
    return trainer


def train_ray(ppo_config, timesteps_total: int = 10):
    analysis = ray.tune.run(
        ppo.PPOTrainer,
        config=ppo_config,
        local_dir=os.path.join(get_project_root(), "models"),
        stop={"timesteps_total": timesteps_total},
        checkpoint_at_end=True,
    )
    return analysis


def load_ray(path, ppo_config):
    """Load ray.

    Load a trained RLlib agent from the specified path.
    Call this before testing a trained agent.

    Args:
        path: Path pointing to the agent's saved checkpoint (only used for RLlib agents)
        ppo_config: dict config

    Returns:
        trainer: RLlib trainer object
    """
    trainer = ppo.PPOTrainer(config=ppo_config)
    trainer.restore(path)
    return trainer


def sample_trainer(trainer, env):
    print("Finished training. Running manual test/inference loop.")
    obs = env.reset()
    done = {"__all__": False}
    # run one iteration until done

    for i in range(10000):
        if done["__all__"]:
            print("game done")
            break
        # get agent from current observation
        agent = list(obs.keys())[0]

        # format observation dict
        print(obs)
        obs = obs[agent]
        env.render()

        # get deterministic action
        # trainer.compute_single_action(obs, policy_id=agent)
        policy = trainer.get_policy(policy_id=agent)
        action_exploration_policy, _, action_info = policy.compute_single_action(obs)
        logits = action_info["action_dist_inputs"]
        action = logits.argmax()
        env.render()
        print("agent ", agent, " action ", action)
        obs, reward, done, _ = env.step({agent: action})
        # observations contain original observations and the action mask
        # print(f"Obs: {obs}, Action: {action}, done: {done}")

    env.render()
    print(env.env.rewards)


def tune_training_loop(timesteps_total=10000):
    """Train trainer and sample."""
    trainer, env, ppo_config = prepare_train()

    # train trainer
    analysis = train_ray(ppo_config, timesteps_total=timesteps_total)
    # reload the checkpoint
    last_chpt_path = analysis._checkpoints[-1]["local_dir"]
    checkpoint_file = glob.glob(
        os.path.join(last_chpt_path, "**", "checkpoint-*"), recursive=True
    )[0]
    trainer_trained = load_ray(checkpoint_file, ppo_config)

    # sample trainer
    sample_trainer(trainer_trained, env)


def manual_training_loop(timesteps_total=10000):
    """Train trainer and sample."""
    trainer, env, ppo_config = prepare_train()
    trainer_trained = train(trainer, max_steps=timesteps_total)

    sample_trainer(trainer_trained, env)


if __name__ == "__main__":
    init(local_mode=True)
    tune_training_loop()



import supersuit as ss
from stable_baselines3 import PPO

from gobblet_rl import gobblet_v1

env = gobblet_v1.parallel_env(render_mode=None, args=None)

env = ss.pettingzoo_env_to_vec_env_v1(env)
env = ss.concat_vec_envs_v1(env, 1, num_cpus=1, base_class="stable_baselines3")

experiment_name = "test"
rollout_fragment_length = 50
seed = 0

model = PPO(
    "MlpPolicy",
    env,
    tensorboard_log=f"/tmp/{experiment_name}",
    verbose=3,
    gamma=0.95,
    n_steps=rollout_fragment_length,
    ent_coef=0.01,
    learning_rate=5e-5,
    vf_coef=1,
    max_grad_norm=0.9,
    gae_lambda=1.0,
    n_epochs=30,
    clip_range=0.3,
    batch_size=150,
    seed=seed,
)
# train_timesteps = 100 * 1000
train_timesteps = 1000
model.learn(total_timesteps=train_timesteps)
model.save(f"policy_gobblet_{train_timesteps}")
